
import numpy
import mixture
from scipy.cluster.vq import whiten

"""
A spectral clustering method.
"""


class SpectralClusterer(object):
    def __init__(self, k, normLaplacian=False):
        """
        Initialise the object. 

        :param k: the number of clusters
        :type k: :class:`int`

        :param normLaplacian: whether to use the normalised Laplacian
        :type normLaplacian: :class:`bool`
        """
        self.k = k
        self.normLaplacian = normLaplacian 

    def cluster(self, graph):
        """
        Take a graph and cluster using spectral methods. The parameter k controls
        how many clusters to look for and normLaplacian specified whether to use
        the normalised Laplacian (as in Shi and Malik (2000)) or the unnormalised
        Laplacian.

        :param graph: the graph to cluster
        :type graph: :class:`apgl.graph.AbstractMatrixGraph`

        :returns:  An array of size graph.getNumVertices() of cluster membership 
        """
        #TODO: Use sparse eigenvalue computations for sparse matrix 
        if self.normLaplacian:
            L = graph.normalisedLaplacianRw()
        else:
            L = graph.laplacianMatrix()


        print(L)

        w, V = numpy.linalg.eig(L)

        sortedInds = numpy.argsort(w)
        sortedInds = sortedInds[0:self.k]
        V = V[:, sortedInds]

        V = whiten(V)

        #Now do some EM clustering
        data = mixture.DataSet()
        data.fromArray(V)
        print(V)

        

        distList = [] 
        for i in range(self.k):
            compList = []
            for j in range(self.k):
                compList.append(mixture.NormalDistribution(2, 0.4))
                    
            comp = mixture.ProductDistribution(compList)

            distList.append(comp)

        numInitialisations = 10
        m = mixture.MixtureModel(self.k, numpy.ones(self.k)/self.k, distList)
        m.randMaxEM(data, numInitialisations, 40, 0.1)

        clust = m.classify(data)

        return clust
