
from apgl.util.Parameter import Parameter
from apgl.util.Util import Util
from apgl.graph.AbstractMatrixGraph import AbstractMatrixGraph
from apgl.generator.AbstractGraphGenerator import AbstractGraphGenerator
import numpy
import math 

class ConfigModelGenerator(AbstractGraphGenerator):
    '''
    A class which generates graphs according to the configuration model. In this model
    one specifies a degree sequence and the generated graph fits the sequence as closely
    as possible. 
    '''
    def __init__(self, degSequence):
        '''
        Create a ConfigModelGenerator object with the given degree sequence.

        :param degSequence: a vector of degrees for each vertex in the graph.
        :type degSequence: :class:`numpy.ndarray`

        '''
        self.setDegSequence(degSequence)

    def setDegSequence(self, degSequence):
        '''
        Set the degree sequence of this object. 

        :param degSequence: a vector of degrees for each vertex in the graph.
        :type degSequence: :class:`numpy.ndarray`
        '''
        Parameter.checkClass(degSequence, numpy.ndarray)
        if degSequence.ndim != 1:
            raise ValueError("Degree sequence must be one dimensional")

        Parameter.checkList(degSequence, Parameter.checkInt, [0, degSequence.shape[0]])

        self.degSequence = degSequence

    def getDegSequence(self):
        """
        :returns: A vector of integers corresponding to the specified degree sequence.
        """
        return self.degSequence 

    def generate(self, graph, requireEmpty=True):
        '''
        Create an Configuration Model graph. Note the the degree sequence given
        in the constructor cannot be guarenteed. The algorithm randomly selects
        two free "spokes" and then tried to connect them. If two vertices are
        already connected the corresponding spokes are not used again. In the case
        that requireEmpty is False then a non-empty graph can be used and the given
        degree sequence is the diffence in degrees between the output graph and
        input one. 

        :param graph: a graph to populate with edges
        :type graph: :class:`apgl.graph.AbstractMatrixGraph`

        :param requireEmpty: if this is set to true then we require an empty graph.
        :type requireEmpty: :class:`bool`

        :returns: The modified input graph. 
        '''
        Parameter.checkClass(graph, AbstractMatrixGraph)
        if requireEmpty and graph.getNumEdges()!= 0:
            raise ValueError("Graph must have no edges")
        if graph.getNumVertices() != self.degSequence.shape[0]:
            raise ValueError("Graph must have same number of vertices as degree sequence")

        expandedInds = Util.expandIntArray(self.degSequence)
        numpy.random.shuffle(expandedInds)

        for i in range(0, len(expandedInds), 2):
            if i != len(expandedInds)-1:
                graph.addEdge(expandedInds[i], expandedInds[i+1])

        return graph

