"""
Some methods which are useful but didn't quite fit into the graph classes. 

"""
import logging 
import numpy 

class GraphUtils(object):
    def __init__(self):
        pass

    @staticmethod 
    def vertexLabelPairs(graph, edgeArray):
        """
        Create an array of pairs of vertex labels for each edge in the graph. Returns
        a tuple of this array and a vector of corresponding edge weights, for the
        indices of the full edge index array as given by getAllEdges.

        :param edgeArray: A numpy array with 2 columns, and values corresponding to vertex indices.
        :type edgeArray: :class:`numpy.ndarray`

        """
        #Parameter.checkList(edgeArray, Parameter.checkIndex, (0, self.getNumVertices()))

        numFeatures = graph.getVertexList().getNumFeatures()
        X = numpy.zeros((edgeArray.shape[0], numFeatures*2))

        for i in range(edgeArray.shape[0]):
            X[i, 0:numFeatures] = graph.getVertex(edgeArray[i, 0])
            X[i, numFeatures:numFeatures*2] = graph.getVertex(edgeArray[i, 1])

        return X 

    @staticmethod
    def vertexLabelExamples(graph):
        """
        Return a set of examples with pairs of vertex labels connected by an
        edge. For undircted graphs there exists an example (v_i, v_j) for
        every (v_j, v_i). Also, there is a set of negative examples where the
        edge does not exist. 
        """
        numFeatures = graph.getVertexList().getNumFeatures()
        numEdges = graph.getNumEdges()

        #Also add non-edges
        logging.info("Computing graph complement")
        cGraph = graph.complement()
        logging.info("Done with " + str(cGraph.getNumEdges()) + " edges.")
        perm = numpy.random.permutation(cGraph.getNumEdges())[0:numEdges]

        X = GraphUtils.vertexLabelPairs(graph, graph.getAllEdges())
        Xc = GraphUtils.vertexLabelPairs(cGraph, cGraph.getAllEdges()[perm, :])
        X = numpy.r_[X, Xc]

        y = numpy.ones(numEdges*2)
        y[numEdges:numEdges*2] = -1
        logging.debug(y)

        #If the graph is undirected add reverse edges
        if graph.isUndirected():
            X2 = numpy.zeros((numEdges*2, numFeatures*2))
            X2[:, 0:numFeatures] = X[:, numFeatures:numFeatures*2]
            X2[:, numFeatures:numFeatures*2] = X[:, 0:numFeatures]
            X = numpy.r_[X, X2]
            y = numpy.r_[y, y]

        return X, y

    @staticmethod
    def treeRoot(treeGraph):
        """
        Find the root of the given tree
        """
        inDegSeq = treeGraph.inDegreeSequence()
        root = numpy.nonzero(inDegSeq==0)[0][0]
        return root 

    @staticmethod 
    def treeDepth(treeGraph):
        """
        Find the depth of the given tree. 
        """
        if not treeGraph.isTree():
            raise ValueError("Input graph is not a tree")
        
        root = GraphUtils.treeRoot(treeGraph)
        distances = treeGraph.dijkstrasAlgorithm(root)
        return numpy.max(distances[distances!=float('inf')])