'''
Created on 23 Jul 2009

@author: charanpal

A wrapper for  LibSVM
'''

from apgl.predictors import * 
from apgl.util.Parameter import Parameter 

import numpy
import re
import scipy
import logging
#import scipy.io.array_import as array_import

#TODO: Create a set() method for each parameter

class LibSVM(AbstractSVMPredictor):
    def __init__(self, kernel='linear', kp=0.1, C=1.0, cost=1.0):
        try:
            import svm
        except ImportError:
            raise 

        self.params = svm.svm_parameter()
        self.params.shrinking = 0
        self.params.cache_size = 500
        self.setKernel(kernel, kp)
        self.setC(C)
        self.setErrorCost(1.0)
        self.setSvmType("C_SVC")

    def setC(self, C):
        Parameter.checkFloat(C, 0.0, float('inf'))
        self.params.C = C

    def setEpsilon(self, epsilon):
        Parameter.checkFloat(epsilon, 0.0, float('inf'))
        self.params.p = epsilon

    def setKernel(self,  kernel='linear', kernelParam=0.1):
        try:
            import svm
        except:
            raise 
        Parameter.checkString(kernel, ["linear", "gaussian", "polynomial"])

        if kernel == "linear":
            kernelType = svm.LINEAR
        elif kernel == "polynomial":
            kernelType = svm.POLY
            if round(kernelParam) != kernelParam:
                raise ValueError("Polynomial degree must be integer:" + str(kernelParam))

            self.params.degree = int(kernelParam)
        elif kernel == "gaussian":
            kernelType = svm.RBF
            self.params.gamma = kernelParam
        else:
            raise ValueError("Invalid kernel :" + str(kernel))

        self.kernelStr = kernel
        self.kp = kernelParam
        self.params.kernel_type = kernelType


    def setSvmType(self, typeStr):
        try:
            import svm
        except:
            raise 

        if typeStr == "C_SVC":
            type = svm.C_SVC
        elif typeStr == "Epsilon_SVR":
            type = svm.EPSILON_SVR
        else:
            raise ValueError("Invalid SVM type :" + str(typeStr))

        self.params.svm_type = type
        self.svmTypeStr = typeStr

    def setErrorCost(self, errorCost):
        """
        The penalty on errors on positive labels. The penalty for negative labels
        is 1.
        """
        Parameter.checkFloat(errorCost, 0.0, float('inf'))

        self.params.nr_weight = 2
        self.params.weight_label = [-1, 1]
        self.params.weight = [1, errorCost]
        #Weird error alert: trying to read self.params.weight results in a crash
        self.errorCost = errorCost

    def setTermination(self, epsilon):
        Parameter.checkFloat(epsilon, 0.0, 1.0)
        self.params.eps = epsilon

    def learnModel(self, X, y):
        try:
            import svm
        except:
            raise 

        if self.params.svm_type == svm.C_SVC and (not ((y == 1) + (y == -1)).all()):
            raise ValueError("Labels must be chosen from {-1, +1} for SVC")
            
        prob = svm.svm_problem(y.tolist(), X.tolist())
        self.model = svm.svm_model(prob, self.params)

    def classify(self, X):
        try:
            import svm
        except:
            raise
        
        if self.params.svm_type != svm.C_SVC:
            raise ValueError("Cannot classify without correct svm_type")

        return self.predict(X, True)

    def predict(self, X, decisionValues=False): 
        """
        Classify a set of examples and return an array of predicted labels
        """
        y = numpy.zeros(X.shape[0])
        
        for i, xi in enumerate(X):
            y[i] = self.model.predict(xi)

        if not decisionValues:
            return y
        else:
            decisions = numpy.zeros(X.shape[0])
            for i, xi in enumerate(X):
                d = self.model.predict_values(xi)
                decisions[i] = d[(-1, 1)]*self.model.labels[0]

            #Make sure the decision values are the correct sign 
            decisions = numpy.abs(decisions)*y

            return y, decisions

    def getWeights(self):
        """
        A terrible but required method to output the SVM weights and value of rho.
        Only works for the linear kernel. 
        """
        try:
            import svm
        except:
            raise

        if self.params.kernel_type != svm.LINEAR:
            raise ValueError("Can only compute weights for linear kernel")

        fileName = "SvmModel"
        self.model.save(fileName)

        file = open(fileName, 'r')
        newLines = []
        startMatrix = False 

        for line in file:
            line = re.sub("\n", "", line)
            outputList = re.split(' ', line)

            if startMatrix:
                line = re.sub("[0-9]+:", "", line) + "\n"
                newLines.append(line)

            if outputList[0] == "rho":
                rho = float(outputList[1])
            elif outputList[0] == "SV":
                startMatrix = True

        file = open(fileName, 'w')
        file.writelines(newLines)

        file = open(fileName, 'r')
        SVs = numpy.loadtxt(file)

        numCols = SVs.shape[1]
        #First column is y times alpha
        #Rest are X

        yAlpha = SVs[:, 0]
        X = SVs[:, 1:numCols]

        weights = numpy.sum(X.T * yAlpha, 1)

        """
        The weights are in reverse, and we have to do this check.
        But model.Label does not exist in the Python object. 
        
        if model.Label(1) == -1
          w = -w;
          b = -b;
        end

        This question from the FAQ is relevant: 

        Why the sign of predicted labels and decision values are sometimes reversed?

        Nothing is wrong. Very likely you have two labels +1/-1 and the first instance
        in your data has -1. Think about the case of labels +5/+10. Since SVM needs to
        use +1/-1, internally we map +5/+10 to +1/-1 according to which label appears
        first. Hence a positive decision value implies that we should predict the
        "internal" +1, which may not be the +1 in the input file.

        """
        if self.model.labels[0] == -1:
            weights = -weights
            rho = -rho

        return weights, -rho

    def loadParams(self, fileName):
        """
        Load a set of SVM parameters from a matlab file. 
        """
        try:
            matDict = scipy.io.loadmat(fileName)
        except IOError:
            raise

        if "C" in list(matDict.keys()):
            self.setC(float(matDict["C"][0]))

        if "kernel" in list(matDict.keys()):
            self.setKernel(str(matDict["kernel"][0]), float(matDict["kernelParamVal"][0])) 

        if "errorCost" in list(matDict.keys()):
            self.setErrorCost(float(matDict["errorCost"][0]))

        if "epsilon" in list(matDict.keys()):
            self.setEpsilon(float(matDict["epsilon"][0]))

        if "svmType" in list(matDict.keys()):
            self.setSvmType(str(matDict["svmType"][0]))

    def saveParams(self, fileName):
        matDict = {}
        matDict["C"] = numpy.array([self.params.C])
        matDict["kernel"] = self.kernelStr
        matDict["kernelParamVal"] = numpy.array([self.kp])
        matDict["epsilon"] = numpy.array([self.params.p])
        matDict["errorCost"] = numpy.array([self.errorCost])
        matDict["svmType"] = self.svmTypeStr

        scipy.io.savemat(fileName, matDict)
        logging.info("Saved file as " + fileName)

    def getC(self):
        return self.params.C

    def getErrorCost(self):
        return(self.errorCost)

    def getKernel(self):
        return self.kernelStr

    def getEpsilon(self):
        return self.params.p

    def getSvmType(self):
        return self.svmTypeStr

    def getTermination(self):
        return self.params.eps

    def getKernelParams(self):
        return self.kp

    def __str__(self):
        return "SvmType=" + self.svmTypeStr +" C="+str(self.params.C)+" kernel="+self.kernelStr+" kp="+str(self.kp)+" epsilon="+str(self.params.p)

    def getModel(self):
        return self.model

    params = None
    model = None