"""
A file to write a faster gaussian kernel
""" 

import numpy
import time 

def KernelGaussian (x1, x2, kp):
    """
    Gaussian kernel
    K(x1,x2,kp) = exp^(-(||x1-x2||)^2 / kp)
    """
    sub = x1 - x2
    norm = (numpy.sum(numpy.abs(sub)**2))**(0.5)
    return numpy.exp(-norm**2 / kp)

def MatrixKernelGaussian(x, kp):
    """
    Create the matrix K
    K[i, j] = KernelGaussian(x[i], x[j], kp)
    """

    K = numpy.empty((x.shape[0], x.shape[0]))

    for i in range(x.shape[0]):
        for j in range(i, x.shape[0]):
            K[i,j] = KernelGaussian(x[i], x[j], kp)
            K[j,i] = K[i,j]
    return K

def MatrixKernelGaussian2(X, kp):
    j1 = numpy.ones((X.shape[0], 1))

    diagK1 = numpy.array([numpy.sum(X**2, 1)])
    K1 = numpy.dot(X, X.T)

    Q = (2*K1 - diagK1*j1.T - j1*diagK1.T)/kp;

    return numpy.exp(Q);

def MatrixKernelGaussian3(X, kp):
    K = numpy.zeros((X.shape[0], X.shape[0]))

    for i in range(X.shape[0]):
        for j in range(X.shape[0]):
            K[i, j] = numpy.exp(-sum((X[i, :]-X[j, :])**2)/kp)
    return K

if __name__ == "__main__":
    numpy.random.seed(0)

    X = numpy.random.rand(5,5)
    kp = 2
    
    K1 = MatrixKernelGaussian(X, kp)
    print(K1)

    K2 = MatrixKernelGaussian2(X, kp)
    print(K2)

    K3 = MatrixKernelGaussian3(X, kp)
    print(K3)

    #Now do some speed tests
    numExamples = 1000
    X = numpy.random.rand(numExamples, numExamples)
    kp = 2

    start = time.clock()
    K1 = MatrixKernelGaussian(X, kp)
    current = time.clock()
    print(("MatrixKernelGaussian took " + str(current-start) + " seconds"))

    start = time.clock()
    K2 = MatrixKernelGaussian2(X, kp)
    current = time.clock()
    print(("MatrixKernelGaussian2 took " + str(current-start) + " seconds"))

    start = time.clock()
    K3 = MatrixKernelGaussian3(X, kp)
    current = time.clock()
    print(("MatrixKernelGaussian3 took " + str(current-start) + " seconds"))