'''
Created on 31 Jul 2009

@author: charanpal
'''

import numpy

class Parameter(object):
    """
    A class of static methods which are useful for type checking. 
    """
    def __init__(self):
        pass

    @staticmethod
    def checkClass(obj, objectType):
        """
        Check if an object belongs to a particular class and raise a ValueError
        if it does not.

        :param objectType: The object to test. `
        """
        if not isinstance(obj, objectType):
            raise ValueError("Expecting object of type " + str(objectType) + " but received " + str(obj.__class__ ))


    @staticmethod
    def checkBoolean(val):
        """
        Check if val is a boolean and raise a ValueError if it is not.

        :param val: The value to test. 
        :type val: :class:`bool`
        """
        if type(val) != bool:
            raise ValueError("Expecting boolean but received " + str(type(val)))

    @staticmethod
    def checkIndex(i, min, max):
        """
        Check if i is an integer value between min inclusive and max exclusive and raise a
        ValueError if it is not. If one requires an open ended range, then min
        or max can be float('inf') for example.

        :param i: The value to test, such that min <= i < max.
        :type i: :class:`int`

        :param min: The minimum value of i.
        :type min: :class:`int`

        :param max: The maximum+1 value of i.
        :type max: :class:`int`
        """
        if not (type(min) in Parameter.intTypes or min == float("-inf")):
            raise ValueError("Minimum value is not an integer: " + str(min))
        if not (type(max) in Parameter.intTypes or max == float("inf")):
            raise ValueError("Maximum value is not an integer: " + str(max))
        if (type(i) not in Parameter.intTypes):
            raise ValueError("Invalid parameter value (not int): " + str(i) + " is " + str(type(i)))

        if i < min or i >= max:
            raise ValueError("Invalid parameter value: " + str(i) + " not in [" + str(min)+ ", " + str(max) + "]" )

    @staticmethod
    def checkInt(i, min, max):
        """
        Check if i is an integer value between min and max inclusive and raise a
        ValueError if it is not. If one requires an open ended range, then min
        or max can be float('inf') for example.

        :param i: The value to test, such that min <= i <= max.
        :type i: :class:`int`

        :param min: The minimum value of i.
        :type min: :class:`int`

        :param max: The maximum value of i.
        :type max: :class:`int`
        """
        if not (type(min) in Parameter.intTypes or min == float("-inf")):
            raise ValueError("Minimum value is not an integer: " + str(min))
        if not (type(max) in Parameter.intTypes or max == float("inf")):
            raise ValueError("Maximum value is not an integer: " + str(max))
        if (type(i) not in Parameter.intTypes):
            raise ValueError("Invalid parameter value (not int): " + str(i) + " is " + str(type(i)))
 
        if i < min or i > max:
            raise ValueError("Invalid parameter value: " + str(i) + " not in [" + str(min)+ ", " + str(max) + "]" )
        
    @staticmethod
    def checkFloat(i, min, max):
        """
        Check if i is an floating point value between min and max inclusive and raise a
        ValueError if it is not. If one requires an open ended range, then min
        or max can be float('inf') for example.

        :param i: The value to test, such that min <= i <= max.
        :type i: :class:`float`

        :param min: The minimum value of i.
        :type min: :class:`float`

        :param max: The maximum value of i.
        :type max: :class:`float`
        """
        if type(min) not in Parameter.floatTypes:
            raise ValueError("Minimum value is not a float: " + str(min))
        if type(max) not in Parameter.floatTypes:
            raise ValueError("Maximum value is not a float: " + str(max))
        if type(i) not in Parameter.floatTypes:
            raise ValueError("Invalid parameter value (not float): " + str(i) + " is " + str(type(i)))

        if i < min or i > max:
            raise ValueError("Invalid parameter value: " + str(i) + " not in [" + str(min)+ ", " + str(max) + "]" )

    @staticmethod
    def checkString(s, strList):
        """
        Check if s is an string in strList and raise a ValueError if it is not.

        :param s: The string to test. 
        :type s: :class:`str`

        :param strList: A list of valid strings.
        :type strList: :class:`list`
        """
        if type(strList) != list:
            raise ValueError("Second parameter is required to be a list: " + str(strList))

        if type(s) != str or s not in strList:
            raise ValueError("Invalid string: " + str(s))

    @staticmethod
    def checkList(lst, func, params):
        """
        Check if a list/ndarray obeys the constaints given by func. For example, in order
        to check if a list/ndarray lst contains integers between 0 and 10, one can use
        Parameter.checkList(lst, Parameter.checkInt, [0, 10]). The first argument
        of checkInt is taken from the input list and the remaining ones correspond to 
        the 3rd parameter of checkList. If an array is used as input, then it must be 1D. 

        :param lst: The list to test.
        :type lst: :class:`lst`

        :param func: A function which raises an exception when an invalid list entry is encountered, otherwise does nothing.
        """
        if type(lst) != list and type(lst) != numpy.ndarray:
            raise ValueError("First parameter is required to be a list or array, not " + str(type(lst)))
        if type(lst) == numpy.ndarray and lst.ndim != 1:
            raise ValueError("First parameter must be 1 dimensional: " + str(lst.shape))

        for i in lst:
            func(i, *params)

    #Note that there are two types of int32 for some weird reason. 
    a = numpy.int8(1)
    b = numpy.int32(2)
    intTypes = [int, numpy.int8, numpy.int16, type(a+b), type(b+a), numpy.int64]
    floatTypes = [float, numpy.float32, numpy.float64]