import logging
import sys
import gc 
import matplotlib
import numpy
import os.path
matplotlib.use('Qt4Agg') # do this before importing pylab
import matplotlib.pyplot as plt
from datetime import date
from apgl.util.PathDefaults import PathDefaults
from apgl.util.DateUtils import DateUtils
from apgl.util.Latex import Latex
from apgl.util.Util import Util
from apgl.graph import * 
from apgl.viroscopy.HIVGraphReader import HIVGraphReader
from apgl.viroscopy.HIVGraphStatistics import HIVGraphStatistics

"""
This script computes some basic statistics on the growing graph. We currently
combine both infection and detection graphs (which are both directed) and hence
look at the contact graph. 
"""
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
numpy.set_printoptions(suppress=True, linewidth=60)

hivReader = HIVGraphReader()
graph = hivReader.readHIVGraph()
fInds = hivReader.getIndicatorFeatureIndices()

#The set of edges indexed by zeros is the contact graph
#The ones indexed by 1 is the infection graph
edgeTypeIndex1 = 0
edgeTypeIndex2 = 1
sGraphContact = graph.getSparseGraph(edgeTypeIndex1)
sGraphInfect = graph.getSparseGraph(edgeTypeIndex2)
sGraphContact = sGraphContact.union(sGraphInfect)
sGraph = sGraphContact
#sGraph = sGraph.subgraph(range(0, 200))

figureDir = PathDefaults.getOutputDir() + "viroscopy/figures/contact/"
resultsDir = PathDefaults.getOutputDir() + "viroscopy/"

graphStats = GraphStatistics()
statsArray = graphStats.scalarStatistics(sGraph, False)
slowStats = True
saveResults = False

logging.info(sGraph)
logging.info("Number of vertices: " + str(statsArray[graphStats.numVerticesIndex]))
logging.info("Number of features: " + str(sGraph.getVertexList().getNumFeatures()))
logging.info("Number of edges: " + str(statsArray[graphStats.numEdgesIndex]))
logging.info("Largest component is " + str(statsArray[graphStats.maxComponentSizeIndex]))
logging.info("Number of components " + str(statsArray[graphStats.numComponentsIndex]))

#sGraph = sGraph.subgraph(components[componentIndex])
vertexArray = sGraph.getVertexList().getVertices()
logging.info("Size of graph we will use: " + str(sGraph.getNumVertices()))

#Some indices
dobIndex = fInds["birthDate"]
detectionIndex = fInds["detectDate"]
deathIndex = fInds["deathDate"]
genderIndex = fInds["gender"]
orientationIndex = fInds["orient"]

contactIndex = fInds["contactTrace"]
donorIndex = fInds["donor"]
randomTestIndex = fInds["randomTest"]
stdIndex = fInds["STD"]
prisonerIndex = fInds["prisoner"]
doctorIndex = fInds["recommendVisit"]

#The 5 most popular provinces
havanaIndex = fInds["CH"]
villaClaraIndex = fInds["VC"]
pinarIndex = fInds["PR"]
holguinIndex = fInds["HO"]
habanaIndex = fInds["LH"]
sanctiIndex = fInds["SS"]

numContactsIndex = fInds["numContacts"]
numTestedIndex = fInds["numTested"]
numPositiveIndex = fInds["numPositive"]

ages = vertexArray[:, detectionIndex] - vertexArray[:, dobIndex]
deaths = vertexArray[:, deathIndex] - vertexArray[:, detectionIndex]
detections = vertexArray[:, detectionIndex]
meanAgeAtDetection = numpy.mean(ages)
stdAgeAtDetection = numpy.std(ages)
meanDeathTime = numpy.mean(deaths)
stdDeathTime = numpy.std(deaths)

#print(numpy.sum(vertexArray[:, 2:22], 0))

startYear = 1900
daysInYear = 365
daysInMonth = 30 
logging.info("Mean age at detection: " + str(meanAgeAtDetection/daysInYear))
logging.info("Std age at detection: " + str(stdAgeAtDetection/daysInYear))

q = 0.9

#Properties of vertex values 
detectionAges = []
deathAfterInfectAges = []
deathAges = []
homoMeans = []
maleSums = []
femaleSums = []
heteroSums = []
biSums = []

contactSums = []
nonContactSums = [] 
donorSums = []
randomTestSums = []
stdSums = []
prisonerSums = []
recommendSums = []

havanaSums = []
villaClaraSums = []
pinarSums = []
holguinSums = []
habanaSums = []

numContactSums = []
numTestedSums = []
numPositiveSums = [] 

monthStep = 3

#Make sure we include all detections 
dayList = range(int(numpy.min(detections)), int(numpy.max(detections)), daysInMonth*monthStep)
dayList.append(numpy.max(detections))

#Day 0 is the start
absDayList = [float(i-numpy.min(detections)) for i in dayList]
subgraphIndicesList = []

#Locations and labels for years
locs = list(range(0, int(absDayList[-1]), daysInYear*2))
labels = numpy.arange(1986, 2006, 2)

logging.info(dayList)
logging.info(absDayList)
plotInd = 1
plotStyles = ['ko-', 'kx-', 'k+-', 'k.-', 'k*-']
plotStyles2 = ['k-', 'r-', 'g-', 'b-', 'c-', 'm-']
plotStyles3 = ['k-', 'k--', 'k-.', 'k:']

#First, compute a series of properties based on vertex values
logging.info("Computing vertex stats")

for i in dayList:
    logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
    subgraphIndices = numpy.nonzero(detections <= i)[0]
    subgraphIndicesList.append(subgraphIndices)
    subgraph = sGraph.subgraph(subgraphIndices)

    #Values based on the properties of the vertices
    subgraphVertexArray = subgraph.getVertexList().getVertices(list(range(0, subgraph.getNumVertices())))
    detectionAges.append(numpy.mean((subgraphVertexArray[:, detectionIndex] - subgraphVertexArray[:, dobIndex]))/daysInYear)
    deathAfterInfectAges.append((numpy.mean(subgraphVertexArray[:, deathIndex] - subgraphVertexArray[:, detectionIndex]))/daysInYear)
    deathAges.append(numpy.mean((subgraphVertexArray[:, deathIndex] - subgraphVertexArray[:, dobIndex]))/daysInYear)
    homoMeans.append(numpy.mean(subgraphVertexArray[:, orientationIndex]))

    nonContactSums.append(subgraphVertexArray.shape[0] - numpy.sum(subgraphVertexArray[:, contactIndex]))
    contactSums.append(numpy.sum(subgraphVertexArray[:, contactIndex]))
    donorSums.append(numpy.sum(subgraphVertexArray[:, donorIndex]))
    randomTestSums.append(numpy.sum(subgraphVertexArray[:, randomTestIndex]))
    stdSums.append(numpy.sum(subgraphVertexArray[:, stdIndex]))
    prisonerSums.append(numpy.sum(subgraphVertexArray[:, prisonerIndex]))
    recommendSums.append(numpy.sum(subgraphVertexArray[:, doctorIndex]))

    heteroSums.append(numpy.sum(subgraphVertexArray[:, orientationIndex]==0))
    biSums.append(numpy.sum(subgraphVertexArray[:, orientationIndex]==1))

    femaleSums.append(numpy.sum(subgraphVertexArray[:, genderIndex]==1))
    maleSums.append(numpy.sum(subgraphVertexArray[:, genderIndex]==0))

    havanaSums.append(numpy.sum(subgraphVertexArray[:, havanaIndex]==1))
    villaClaraSums.append(numpy.sum(subgraphVertexArray[:, villaClaraIndex]==1))
    pinarSums.append(numpy.sum(subgraphVertexArray[:, pinarIndex]==1))
    holguinSums.append(numpy.sum(subgraphVertexArray[:, holguinIndex]==1))
    habanaSums.append(numpy.sum(subgraphVertexArray[:, habanaIndex]==1))

    numContactSums.append(numpy.mean(subgraphVertexArray[:, numContactsIndex]))
    numTestedSums.append(numpy.mean(subgraphVertexArray[:, numTestedIndex]))
    numPositiveSums.append(numpy.mean(subgraphVertexArray[:, numPositiveIndex]))

def plotVertexStats():
    global plotInd 

    plt.figure(plotInd)
    plt.plot(absDayList, detectionAges)
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detection Age (years)")
    plt.savefig(figureDir + "DetectionMeansGrowth.eps")
    plotInd += 1

    plt.figure(plotInd)
    plt.plot(absDayList, heteroSums, 'k-', absDayList, biSums, 'k--', absDayList, femaleSums, 'k-.', absDayList, maleSums, 'k:')
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Heterosexual", "MSM", "Female", "Male"), loc="upper left")
    plt.savefig(figureDir + "OrientationGenderGrowth.eps")
    plotInd += 1

    """
    plt.figure(plotInd)
    plt.plot(absDayList, femaleSums, 'r', absDayList, maleSums, 'b')
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Female", "Male"), loc="upper left")
    plt.savefig(figureDir + "GenderGrowth.eps")
    plotInd += 1
    """
    #Plot of detection types
    plt.figure(plotInd)
    plt.plot(absDayList, contactSums, plotStyles2[0], absDayList, donorSums, plotStyles2[1], absDayList, randomTestSums, plotStyles2[2], absDayList, stdSums, plotStyles2[3], absDayList, prisonerSums, plotStyles2[4], absDayList, recommendSums, plotStyles2[5])
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Contact tracing", "Blood donation", "Random test", "STD", "Other test", "Doctor recommendation"), loc="upper left")
    plt.savefig(figureDir + "DetectionGrowth.eps")
    plotInd += 1

    plt.figure(plotInd)
    plt.plot(absDayList, numContactSums, plotStyles3[0], absDayList, numTestedSums, plotStyles3[1], absDayList, numPositiveSums, plotStyles3[2])
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Number")
    plt.legend(("No. Contacts", "No. Tested", "No. Positive"), loc="center left")
    plt.savefig(figureDir + "ContactsGrowth.eps")
    plotInd += 1

    plt.figure(plotInd)
    plt.plot(absDayList, havanaSums, plotStyles2[0], absDayList, villaClaraSums, plotStyles2[1], absDayList, pinarSums, plotStyles2[2], absDayList, holguinSums, plotStyles2[3], absDayList, habanaSums, plotStyles2[4])
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Detections")
    plt.legend(("Havana City", "Villa Clara", "Pinar del Rio", "Holguin", "La Habana"), loc="upper left")
    plt.savefig(figureDir + "ProvinceGrowth.eps")
    plotInd += 1

def computeConfigScalarStats():
    logging.info("Computing configuration model scalar stats")

    numGraphs = 5
    graphFileNameBase = resultsDir + "ConfigGraph"
    resultsFileNameBase = resultsDir + "ContactGraphScalarStats"
    #graphStats.useFloydWarshall = True

    for j in range(numGraphs):
        resultsFileName = resultsFileNameBase + str(j)

        if not os.path.isfile(resultsFileName):
            configGraph = SparseGraph.load(graphFileNameBase + str(j))
            statsArray = graphStats.sequenceScalarStats(configGraph, subgraphIndicesList, slowStats)
            Util.savePickle(statsArray, resultsFileName, True)
            gc.collect()

    logging.info("All done")

def computeConfigVectorStats():
    logging.info("Computing configuration model vector stats")

    numGraphs = 5
    graphFileNameBase = resultsDir + "ConfigGraph"
    resultsFileNameBase = resultsDir + "ContactGraphVectorStats"
    #graphStats.useFloydWarshall = True

    for j in range(numGraphs):
        resultsFileName = resultsFileNameBase + str(j)

        if not os.path.isfile(resultsFileName):
            configGraph = SparseGraph.load(graphFileNameBase + str(j))
            statsDictList = graphStats.sequenceVectorStats(configGraph, subgraphIndicesList)
            Util.savePickle(statsDictList, resultsFileName, False)
            gc.collect()

    logging.info("All done")

def plotScalarStats():
    logging.info("Computing scalar stats")

    resultsFileName = resultsDir + "ContactGrowthScalarStats.pkl"

    if saveResults:
        statsArray = graphStats.sequenceScalarStats(sGraph, subgraphIndicesList, slowStats)
        Util.savePickle(statsArray, resultsFileName, True)

        #Now compute statistics on the configuration graphs 
    else:
        statsArray = Util.loadPickle(resultsFileName)

        global plotInd

        #Output all the results into plots
        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.maxComponentSizeIndex], 'k-', absDayList, statsArray[:, graphStats.maxComponentEdgesIndex], 'k--', absDayList, statsArray[:, graphStats.secondComponentSizeIndex], 'k:')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Num vertices/edges")
        plt.legend(("Max Comp. Vertices", "Max Comp. Edges", "2nd Comp. Vertices"), loc="upper left")
        plt.savefig(figureDir + "MaxComponentSizeGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.numComponentsIndex], 'k-', absDayList, statsArray[:, graphStats.numNonSingletonComponentsIndex], 'k--', absDayList, statsArray[:, graphStats.numTriOrMoreComponentsIndex], 'k:')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Number of components")
        plt.legend(("Size >= 1", "Size >= 2", "Size >= 3"), loc="upper left")
        plt.savefig(figureDir + "NumComponentsGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.meanComponentSizeIndex], 'k-')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Mean component size")
        plt.savefig(figureDir + "MeanComponentSizeGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.diameterIndex], 'k-')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component diameter")
        plt.savefig(figureDir + "MaxComponentDiameterGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.effectiveDiameterIndex], 'k-')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component effective diameter")
        plt.savefig(figureDir + "MaxComponentEffDiameterGrowth.eps")
        plotInd += 1

        #Make sure we don't include 0 in the array
        vertexIndex = numpy.argmax(statsArray[:, graphStats.numVerticesIndex] > 0)
        edgeIndex = numpy.argmax(statsArray[:, graphStats.numEdgesIndex] > 0)
        minIndex = numpy.maximum(vertexIndex, edgeIndex)

        plt.figure(plotInd)
        plt.plot(numpy.log(statsArray[minIndex:, graphStats.numVerticesIndex]), numpy.log(statsArray[minIndex:, graphStats.numEdgesIndex]), 'k-', numpy.log(statsArray[minIndex:, graphStats.maxComponentSizeIndex]), numpy.log(statsArray[minIndex:, graphStats.maxComponentEdgesIndex]), 'k-.')
        plt.xlabel("log(No. vertices)")
        plt.ylabel("log(No. edges)")
        plt.legend(("All vertices", "Max component"), loc="upper left")
        plt.savefig(figureDir + "LogVerticesEdgesGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.meanDegreeIndex], 'k-', absDayList, statsArray[:, graphStats.maxCompMeanDegreeIndex], 'k-.')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Mean degree")
        plt.legend(("All vertices", "Max component"), loc="lower right")
        plt.savefig(figureDir + "MeanDegrees.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.densityIndex], 'k')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Density")
        plt.savefig(figureDir + "DensityGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.powerLawIndex], 'k-')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Alpha")
        plt.savefig(figureDir + "PowerLawGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.geodesicDistanceIndex], 'k')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Geodesic distance")
        plt.savefig(figureDir + "GeodesicGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.harmonicGeoDistanceIndex], 'k')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Mean harmonic geodesic distance")
        plt.savefig(figureDir + "HarmonicGeodesicGrowth.eps")
        plotInd += 1

        #print(statsArray[:, graphStats.harmonicGeoDistanceIndex])

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.geodesicDistMaxCompIndex], 'k')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Max component geodesic distance")
        plt.savefig(figureDir + "MaxCompGeodesicGrowth.eps")
        plotInd += 1

        #Print out the mean geodesic distances for Hector
        
        #print(statsArray[:, graphStats.geodesicDistMaxCompIndex][1:])


def plotVectorStats():
    #Finally, compute some vector stats at various points in the graph
    logging.info("Computing vector stats")
    dayList2 = [DateUtils.getDayDelta(date(1989, 12, 31), startYear)]
    dayList2.append(DateUtils.getDayDelta(date(1993, 12, 31), startYear))
    dayList2.append(DateUtils.getDayDelta(date(1997, 12, 31), startYear))
    dayList2.append(DateUtils.getDayDelta(date(2001, 12, 31), startYear))
    dayList2.append(int(numpy.max(detections)))
    
    subgraphIndicesList = []
    global plotInd
    
    for i in dayList2:
        logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
        subgraphIndices = numpy.nonzero(detections <= i)[0]
        subgraphIndicesList.append(subgraphIndices)

    resultsFileName = resultsDir + "ContactGrowthVectorStats.pkl"

    if saveResults:
        statsDictList = graphStats.sequenceVectorStats(sGraph, subgraphIndicesList)
        Util.savePickle(statsDictList, resultsFileName, False)
    else:
        statsDictList = Util.loadPickle(resultsFileName)

        triangleDistArray = numpy.zeros((len(dayList2), 100))
        hopPlotArray = numpy.zeros((len(dayList2), 27))
        componentsDistArray = numpy.zeros((len(dayList2), 3000))
        numVerticesEdgesArray = numpy.zeros((len(dayList2), 2), numpy.int)
        numVerticesEdgesArray[:, 0] = [len(sgl) for sgl in subgraphIndicesList]
        numVerticesEdgesArray[:, 1] = [sGraph.subgraph(sgl).getNumEdges() for sgl in subgraphIndicesList]

        binWidths = numpy.arange(0, 0.50, 0.05)
        eigVectorDists = numpy.zeros((len(dayList2), binWidths.shape[0]-1), numpy.int)

        femaleSums = numpy.zeros(len(dayList2))
        maleSums = numpy.zeros(len(dayList2))
        heteroSums = numpy.zeros(len(dayList2))
        biSums = numpy.zeros(len(dayList2))

        contactSums = numpy.zeros(len(dayList2))
        nonContactSums = numpy.zeros(len(dayList2))
        donorSums = numpy.zeros(len(dayList2))
        randomTestSums = numpy.zeros(len(dayList2))
        stdSums = numpy.zeros(len(dayList2))
        prisonerSums = numpy.zeros(len(dayList2))
        recommendSums = numpy.zeros(len(dayList2))
        
        meanAges = numpy.zeros(len(dayList2))
        degrees = numpy.zeros((len(dayList2), 20))

        provinces = numpy.zeros((len(dayList2), 15))

        havanaSums = numpy.zeros(len(dayList2))
        villaClaraSums = numpy.zeros(len(dayList2))
        pinarSums = numpy.zeros(len(dayList2))
        holguinSums = numpy.zeros(len(dayList2))
        habanaSums = numpy.zeros(len(dayList2))
        sanctiSums = numpy.zeros(len(dayList2))

        #Note that death has a lot of missing values
        for j in range(len(dayList2)):
            dateStr = (str(DateUtils.getDateStrFromDay(dayList2[j], startYear)))
            logging.info(dateStr)
            statsDict = statsDictList[j]

            degreeDist = statsDict["outDegreeDist"]
            degreeDist = degreeDist/float(numpy.sum(degreeDist))

            eigenDist = statsDict["eigenDist"]
            eigenDist = numpy.log(eigenDist[eigenDist>=10**-1])

            hopCount = statsDict["hopCount"]
            hopCount = numpy.log10(hopCount)
            #hopCount = numpy.array(hopCount, numpy.float64)/numpy.max(hopCount)
            hopPlotArray[j, 0:hopCount.shape[0]] = hopCount

            triangleDist = statsDict["triangleDist"]
            triangleDist = numpy.array(triangleDist, numpy.float64)/numpy.sum(triangleDist)
            triangleDistArray[j, 0:triangleDist.shape[0]] = triangleDist

            maxEigVector = statsDict["maxEigVector"]
            eigenvectorInds = numpy.flipud(numpy.argsort(numpy.abs(maxEigVector)))
            top10eigenvectorInds = eigenvectorInds[0:numpy.round(eigenvectorInds.shape[0]/10.0)]
            maxEigVector = numpy.abs(maxEigVector[eigenvectorInds])
            #print(maxEigVector)
            eigVectorDists[j, :] = numpy.histogram(maxEigVector, binWidths)[0]

            componentsDist = statsDict["componentsDist"]
            componentsDist = numpy.array(componentsDist, numpy.float64)/numpy.sum(componentsDist)
            componentsDistArray[j, 0:componentsDist.shape[0]] = componentsDist

            plotInd2 = plotInd

            plt.figure(plotInd2)
            plt.plot(numpy.arange(degreeDist.shape[0]), degreeDist, plotStyles2[j], label=dateStr)
            plt.xlabel("Degree")
            plt.ylabel("Probability")
            plt.ylim((0, 0.5))
            plt.savefig(figureDir + "DegreeDist" +  ".eps")
            plt.legend()
            plotInd2 += 1

            """
            plt.figure(plotInd2)
            plt.plot(numpy.arange(eigenDist.shape[0]), eigenDist, label=dateStr)
            plt.xlabel("Eigenvalue rank")
            plt.ylabel("log(Eigenvalue)")
            plt.savefig(figureDir + "EigenDist" +  ".eps")
            plt.legend()
            plotInd2 += 1
            """

            #How does kleinberg do the hop plots 
            plt.figure(plotInd2)
            plt.plot(numpy.arange(hopCount.shape[0]), hopCount, plotStyles[j], label=dateStr)
            plt.xlabel("k")
            plt.ylabel("log10(pairs)")
            plt.ylim( (2.5, 7) )
            plt.legend(loc="lower right")
            plt.savefig(figureDir + "HopCount" + ".eps")
            plotInd2 += 1
            
            plt.figure(plotInd2)
            plt.plot(numpy.arange(maxEigVector.shape[0]), maxEigVector, plotStyles2[j], label=dateStr)
            plt.xlabel("Rank")
            plt.ylabel("log(eigenvector coefficient)")
            plt.savefig(figureDir + "MaxEigVector" +  ".eps")
            plt.legend()
            plotInd2 += 1

            #Compute some information the 10% most central vertices
            
            subgraphIndices = numpy.nonzero(detections <= dayList2[j])[0]
            subgraph = sGraph.subgraph(subgraphIndices)
            subgraphVertexArray = subgraph.getVertexList().getVertices()

            femaleSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, genderIndex]==1)
            maleSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, genderIndex]==0)
            heteroSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, orientationIndex]==0)
            biSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, orientationIndex]==1)

            contactSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, contactIndex])
            donorSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, donorIndex])
            randomTestSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, randomTestIndex])
            stdSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, stdIndex])
            prisonerSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, prisonerIndex])
            recommendSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, doctorIndex])

            meanAges[j] = numpy.mean(subgraphVertexArray[top10eigenvectorInds, detectionIndex] - subgraphVertexArray[top10eigenvectorInds, dobIndex])/daysInYear

            havanaSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, havanaIndex])
            villaClaraSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, villaClaraIndex])
            pinarSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, pinarIndex])
            holguinSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, holguinIndex])
            habanaSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, habanaIndex])
            sanctiSums[j] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, sanctiIndex])

            provinces[j, :] = numpy.sum(subgraphVertexArray[top10eigenvectorInds, 22:37], 0)

            ddist = numpy.bincount(subgraph.outDegreeSequence()[top10eigenvectorInds])
            degrees[j, 0:ddist.shape[0]] = numpy.array(ddist, numpy.float)/numpy.sum(ddist)

            plt.figure(plotInd2)
            plt.plot(numpy.arange(degrees[j, :].shape[0]), degrees[j, :], plotStyles2[j], label=dateStr)
            plt.xlabel("Degree")
            plt.ylabel("Probability")
            #plt.ylim((0, 0.5))
            plt.savefig(figureDir + "DegreeDistCentral" +  ".eps")
            plt.legend()
            plotInd2 += 1

        precision = 4
        dateStrList = [DateUtils.getDateStrFromDay(day, startYear) for day in dayList2]

        print("Hop counts")
        print((Latex.listToRow(dateStrList)))
        print(Latex.array2DToRows(hopPlotArray.T))

        print("Edges and vertcies")
        print((Latex.listToRow(dateStrList)))
        print((Latex.array2DToRows(numVerticesEdgesArray.T, precision)))

        print("\n\nEigenvector distribution")
        print((Latex.array1DToRow(binWidths[1:]) + "\\\\"))
        print((Latex.array2DToRows(eigVectorDists)))

        print("\n\nDistribution of component sizes")
        componentsDistArray = componentsDistArray[:, 0:componentsDist.shape[0]]
        nonZeroCols = numpy.sum(componentsDistArray, 0)!=0
        componentsDistArray = numpy.r_[numpy.array([numpy.arange(componentsDistArray.shape[1])[nonZeroCols]]), componentsDistArray[:, nonZeroCols]]
        print((Latex.listToRow(dateStrList)))
        print((Latex.array2DToRows(componentsDistArray.T, precision)))

        print("\n\nDistribution of triangle participations")
        triangleDistArray = triangleDistArray[:, 0:triangleDist.shape[0]]
        nonZeroCols = numpy.sum(triangleDistArray, 0)!=0
        triangleDistArray = numpy.r_[numpy.array([numpy.arange(triangleDistArray.shape[1])[nonZeroCols]])/2, triangleDistArray[:, nonZeroCols]]
        print((Latex.listToRow(dateStrList)))
        print((Latex.array2DToRows(triangleDistArray. T, precision)))

        print("\n\nDistribution of central vertices")
        print((Latex.listToRow(dateStrList)))
        subgraphSizes = numpy.array(maleSums + femaleSums, numpy.float)
        print("Female & " + Latex.array1DToRow(femaleSums/subgraphSizes) + "\\\\")
        print("Male & " + Latex.array1DToRow(maleSums/subgraphSizes) + "\\\\")
        print("\hline")
        print("Heterosexual & " + Latex.array1DToRow(heteroSums/subgraphSizes) + "\\\\")
        print("Bisexual & " + Latex.array1DToRow(biSums/subgraphSizes) + "\\\\")
        print("\hline")
        print("Contact traced & " + Latex.array1DToRow(contactSums/subgraphSizes) + "\\\\")
        print("Blood donor & " + Latex.array1DToRow(donorSums/subgraphSizes) + "\\\\")
        print("RandomTest & " + Latex.array1DToRow(randomTestSums/subgraphSizes) + "\\\\")
        print("STD & " + Latex.array1DToRow(stdSums/subgraphSizes) + "\\\\")
        print("Prisoner & " + Latex.array1DToRow(prisonerSums/subgraphSizes) + "\\\\")
        print("Doctor recommendation & " + Latex.array1DToRow(recommendSums/subgraphSizes) + "\\\\")
        print("\hline")
        print("Mean ages & " + Latex.array1DToRow(meanAges) + "\\\\")
        print("\hline")
        print("Holguin & " + Latex.array1DToRow(holguinSums/subgraphSizes) + "\\\\")
        print("La Habana & " + Latex.array1DToRow(habanaSums/subgraphSizes) + "\\\\")
        print("Havana City & " + Latex.array1DToRow(havanaSums/subgraphSizes) + "\\\\")
        print("Pinar del Rio & " + Latex.array1DToRow(pinarSums/subgraphSizes) + "\\\\")
        print("Sancti Spiritus & " + Latex.array1DToRow(sanctiSums/subgraphSizes) + "\\\\")
        print("Villa Clara & " + Latex.array1DToRow(villaClaraSums/subgraphSizes) + "\\\\")
        
        print("\n\nProvinces")
        print(Latex.array2DToRows(provinces))

        print("\n\nDegree distribution")
        print(Latex.array2DToRows(degrees))



def plotOtherStats():
    #Let's look at geodesic distances in subgraphs and communities
    logging.info("Computing other stats")

    resultsFileName = resultsDir + "ContactGrowthOtherStats.pkl"
    hivGraphStats = HIVGraphStatistics(fInds)

    if saveResults:
        statsArray = hivGraphStats.sequenceScalarStats(sGraph, subgraphIndicesList)
        #statsArray["dayList"] = absDayList
        Util.savePickle(statsArray, resultsFileName, True)
    else:
        statsArray = Util.loadPickle(resultsFileName)
        #Just load the harmonic geodesic distances of the full graph 
        resultsFileName = resultsDir + "ContactGrowthScalarStats.pkl"
        statsArray2 = Util.loadPickle(resultsFileName)


        global plotInd

        msmGeodesic = statsArray[:, hivGraphStats.msmGeodesicIndex]
        msmGeodesic[msmGeodesic < 0] = 0
        msmGeodesic[msmGeodesic == float('inf')] = 0

        #Output all the results into plots
        plt.figure(plotInd)
        plt.plot(absDayList, msmGeodesic, 'k-', absDayList, statsArray[:, hivGraphStats.mostConnectedGeodesicIndex], 'k--')
        plt.xticks(locs, labels)
        #plt.ylim([0, 0.1])
        plt.xlabel("Year")
        plt.ylabel("Mean harmonic geodesic distance")
        plt.legend(("MSM individuals", "Top 10% degree"), loc="upper right")
        plt.savefig(figureDir + "MSM10Geodesic" + ".eps")
        plotInd += 1


        plt.figure(plotInd)
        plt.plot(absDayList, statsArray2[:, graphStats.harmonicGeoDistanceIndex], 'k-', absDayList, statsArray[:, hivGraphStats.menSubgraphGeodesicIndex], 'k--')
        plt.xticks(locs, labels)
        plt.ylim([0, 200.0])
        plt.xlabel("Year")
        plt.ylabel("Mean harmonic geodesic distance")
        plt.legend(("All individuals", "Men subgraph"), loc="upper right")
        plt.savefig(figureDir + "MenSubgraphGeodesic" + ".eps")
        plotInd += 1




#plotVertexStats()
#plotScalarStats()
#plotVectorStats()
#plotOtherStats()
#plt.show()

#computeConfigScalarStats()
computeConfigVectorStats()

"""
Probability of adding node based on degree - try to find how we can generate data.
Mean Time between first and last infection for each person
"""