import logging
import sys
import matplotlib
import numpy
from datetime import date
matplotlib.use('WXAgg') # do this before importing pylab
import matplotlib.pyplot as plt
from apgl.util.PathDefaults import PathDefaults
from apgl.util.DateUtils import DateUtils
from apgl.util.Latex import Latex
from apgl.util.Util import Util
from apgl.graph import *
from apgl.viroscopy.HIVGraphReader import HIVGraphReader

"""
This script computes some basic statistics on the growing infection graph.
"""

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

undirected = False 
hivReader = HIVGraphReader()
graph = hivReader.readHIVGraph(undirected, indicators=False)
fInds = hivReader.getNonIndicatorFeatureIndices()


figureDir = PathDefaults.getOutputDir() + "viroscopy/figures/infect/"
resultsDir = PathDefaults.getOutputDir() + "viroscopy/"

#The set of edges indexed by zeros is the contact graph
#The ones indexed by 1 is the infection graph
edgeTypeIndex1 = 0
edgeTypeIndex2 = 1
sGraphContact = graph.getSparseGraph(edgeTypeIndex1)
sGraphInfect = graph.getSparseGraph(edgeTypeIndex2)

sGraph = sGraphInfect
#sGraph = sGraph.subgraph(range(0, 500))

graphStats = GraphStatistics()
statsArray = graphStats.scalarStatistics(sGraph, False)
slowStats = True
saveResults = False

logging.info(sGraph)
logging.info("Number of vertices: " + str(statsArray[graphStats.numVerticesIndex]))
logging.info("Number of features: " + str(sGraph.getVertexList().getNumFeatures()))
logging.info("Number of edges: " + str(statsArray[graphStats.numEdgesIndex]))
logging.info("Largest component is " + str(statsArray[graphStats.maxComponentSizeIndex]))
logging.info("Number of components " + str(statsArray[graphStats.numComponentsIndex]))

#sGraph = sGraph.subgraph(components[componentIndex])
vertexArray = sGraph.getVertexList().getVertices(list(range(0, sGraph.getVertexList().getNumVertices())))
logging.info("Size of graph we will use: " + str(sGraph.getNumVertices()))

#Some indices
dobIndex = fInds["birthDate"]
detectionIndex = fInds["detectDate"]
deathIndex = fInds["deathDate"]
genderIndex = fInds["gender"]
orientationIndex = fInds["orient"]
locationIndex = fInds["province"]

detections = vertexArray[:, detectionIndex]

startYear = 1900
daysInYear = 365
daysInMonth = 30

q = 0.9

#This is a set of days to record simple statistics 
monthStep = 3
dayList = list(range(int(numpy.min(detections)), int(numpy.max(detections)), daysInMonth*monthStep))
dayList.append(numpy.max(detections))

absDayList = [float(i-numpy.min(detections)) for i in dayList]
subgraphIndicesList = []

#Locations and labels for years
locs = list(range(0, int(absDayList[-1]), daysInYear*2))
labels = numpy.arange(1986, 2006, 2)

#This is a set of days to record more complex vectorial statistics
monthStep2 = 60
dayList2 = [DateUtils.getDayDelta(date(1989, 12, 31), startYear)]
dayList2.append(DateUtils.getDayDelta(date(1993, 12, 31), startYear))
dayList2.append(DateUtils.getDayDelta(date(1997, 12, 31), startYear))
dayList2.append(DateUtils.getDayDelta(date(2001, 12, 31), startYear))
dayList2.append(int(numpy.max(detections)))
subgraphIndicesList2 = []

logging.info(dayList2)

for i in dayList2:
    logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
    subgraphIndices = numpy.nonzero(detections <= i)[0]
    subgraphIndicesList2.append(subgraphIndices)

logging.info(dayList)
plotInd = 1
plotStyles = ['ko-', 'kx-', 'k+-', 'k.-', 'k*-']
plotStyles2 = ['k-', 'r-', 'g-', 'b-', 'c-', 'm-']
plotStyles3 = ['k-', 'k--', 'k-.', 'k:']

for i in dayList:
    logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
    subgraphIndices = numpy.nonzero(detections <= i)[0]
    subgraphIndicesList.append(subgraphIndices)

#Check some properties of the trees 
def plotTreeStats():
    logging.info("Computing tree stats")
    resultsFileName = resultsDir + "InfectGrowthTreeStats.pkl"

    if saveResults:
        statsDictList = []

        for j in range(len(subgraphIndicesList2)):
            Util.printIteration(j, 1, len(subgraphIndicesList2))
            subgraphIndices = subgraphIndicesList2[j]
            subgraph = sGraph.subgraph(subgraphIndices)
            logging.info("Finding trees")
            trees = subgraph.findTrees()
            logging.info("Computing tree statistics")
            statsDict = {}

            locationEntropy = []
            orientEntropy = []
            detectionRanges = []

            for i in range(len(trees)):
                if len(trees[i]) > 1:
                    treeGraph = subgraph.subgraph(trees[i])
                    vertexArray = treeGraph.getVertexList().getVertices(list(range(treeGraph.getNumVertices())))
                    
                    locationEntropy.append(Util.entropy(vertexArray[:, locationIndex]))
                    orientEntropy.append(Util.entropy(vertexArray[:, orientationIndex]))
                    
                    detections = vertexArray[:, detectionIndex]
                    detectionRanges.append(numpy.max(detections) - numpy.min(detections))

            statsDict["locationEnt"] = numpy.array(locationEntropy)
            statsDict["orientEnt"] = numpy.array(orientEntropy)
            statsDict["detectRanges"] = numpy.array(detectionRanges)
            statsDictList.append(statsDict)

        Util.savePickle(statsDictList, resultsFileName, True)
    else:
        statsDictList = Util.loadPickle(resultsFileName)
        
        locBins = numpy.arange(0, 2.4, 0.2)
        detectBins = numpy.arange(0, 6500, 500)
        locationEntDists = []
        orientEntDists = []
        detectionDists = [] 

        for j in range(0, len(dayList2)):
            dateStr = (str(DateUtils.getDateStrFromDay(dayList2[j], startYear)))
            logging.info(dateStr)
            statsDict = statsDictList[j]
            plotInd2 = plotInd

            locationEntDists.append(statsDict["locationEnt"])
            orientEntDists.append(statsDict["orientEnt"])
            detectionDists.append(statsDict["detectRanges"])

            print((numpy.min(statsDict["detectRanges"])))
            print((numpy.max(statsDict["detectRanges"])))

        dateStrs = [DateUtils.getDateStrFromDay(dayList2[i], startYear) for i in range(1, len(dayList2))]

        plt.figure(plotInd2)
        histOut = plt.hist(locationEntDists, locBins, normed=True)
        plt.xlabel("Location Entropy")
        plt.ylabel("Probability Density")
        plt.savefig(figureDir + "LocationEnt" +  ".eps")
        #plt.legend()
        plotInd2 += 1

        plt.figure(plotInd2)
        histOut = plt.hist(orientEntDists, normed=True)
        plt.xlabel("Orientation Entropy")
        plt.ylabel("Probability Density")
        plt.savefig(figureDir + "OrientEnt" +  ".eps")
        #plt.legend()
        plotInd2 += 1

        plt.figure(plotInd2)
        histOut = plt.hist(detectionDists, detectBins, normed=True)
        plt.xlabel("Detection Range (days)")
        plt.ylabel("Probability Density")
        plt.savefig(figureDir + "DetectionRanges" +  ".eps")
        #plt.legend()
        plotInd2 += 1


def plotScalarStats():
    logging.info("Computing scalar stats")
    resultsFileName = resultsDir + "InfectGrowthScalarStats.pkl"

    if saveResults:
        statsArray = graphStats.sequenceScalarStats(sGraph, subgraphIndicesList, treeStats=True)
        Util.savePickle(statsArray, resultsFileName, True)
    else:
        statsArray = Util.loadPickle(resultsFileName)

        global plotInd

        #Output all the results into plots

        #Make sure we don't include 0 in the array
        vertexIndex = numpy.argmax(statsArray[:, graphStats.numVerticesIndex] > 0)
        edgeIndex = numpy.argmax(statsArray[:, graphStats.numEdgesIndex] > 0)
        minIndex = numpy.maximum(vertexIndex, edgeIndex)

        plt.figure(plotInd)
        plt.plot(numpy.log(statsArray[minIndex:, graphStats.numVerticesIndex]), numpy.log(statsArray[minIndex:, graphStats.numEdgesIndex]))
        plt.xlabel("log(|V|)")
        plt.ylabel("log(|E|)")
        plt.savefig(figureDir + "LogVerticesEdgesGrowth.eps")
        plotInd += 1

        plt.figure(plotInd)
        plt.plot(absDayList, statsArray[:, graphStats.maxTreeSizeIndex], 'k-', absDayList, statsArray[:, graphStats.maxTreeDepthIndex], 'k--', absDayList, statsArray[:, graphStats.secondTreeSizeIndex], 'k-.', absDayList, statsArray[:, graphStats.secondTreeDepthIndex], 'k:')
        plt.xticks(locs, labels)
        plt.xlabel("Year")
        plt.ylabel("Size/Depth")
        plt.legend(("Max Size", "Max Depth", "Second Size", "Second Depth"), loc="upper left")
        plt.savefig(figureDir + "MaxTreeGrowth.eps")
        
        plotInd += 1


def plotVectorStats():
    #Finally, compute some vector stats at various points in the graph
    logging.info("Computing vector stats")
    global plotInd
    resultsFileName = resultsDir + "InfectGrowthVectorStats.pkl"

    if saveResults:
        statsDictList = graphStats.sequenceVectorStats(sGraph, subgraphIndicesList2, True)
        Util.savePickle(statsDictList, resultsFileName, True)
    else:
        statsDictList = Util.loadPickle(resultsFileName)

        treeSizesDistArray = numpy.zeros((len(dayList2), 3000))
        treeDepthsDistArray = numpy.zeros((len(dayList2), 100))
        numVerticesEdgesArray = numpy.zeros((len(dayList2), 2), numpy.int)
        numVerticesEdgesArray[:, 0] = [len(sgl) for sgl in subgraphIndicesList2]
        numVerticesEdgesArray[:, 1] = [sGraph.subgraph(sgl).getNumEdges() for sgl in subgraphIndicesList2]

        for j in range(len(dayList2)):
            dateStr = (str(DateUtils.getDateStrFromDay(dayList2[j], startYear)))
            logging.info(dateStr)
            statsDict = statsDictList[j]

            degreeDist = statsDict["outDegreeDist"]
            degreeDist = degreeDist/float(numpy.sum(degreeDist))

            maxEigVector = statsDict["maxEigVector"]
            maxEigVector = numpy.flipud(numpy.sort(numpy.abs(maxEigVector)))
            maxEigVector = numpy.log(maxEigVector[maxEigVector>0])

            treeSizesDist = statsDict["treeSizesDist"]
            treeSizesDist = numpy.array(treeSizesDist, numpy.float64)/numpy.sum(treeSizesDist)
            treeSizesDistArray[j, 0:treeSizesDist.shape[0]] = treeSizesDist

            treeDepthsDist = statsDict["treeDepthsDist"]
            #print(treeDepthsDist)
            treeDepthsDist = numpy.array(treeDepthsDist, numpy.float64)/numpy.sum(treeDepthsDist)
            treeDepthsDistArray[j, 0:treeDepthsDist.shape[0]] = treeDepthsDist

            plotInd2 = plotInd

            plt.figure(plotInd2)
            plt.plot(numpy.arange(degreeDist.shape[0]), degreeDist, label=dateStr)
            plt.xlabel("Degree")
            plt.ylabel("Probability")
            plt.ylim((0, 0.8))
            plt.legend()
            plt.savefig(figureDir + "DegreeDist" +  ".eps")
            plotInd2 += 1

        dateStrList = [DateUtils.getDateStrFromDay(day, startYear) for day in dayList2]
        precision = 4 

        treeSizesDistArray = treeSizesDistArray[:, 0:treeSizesDist.shape[0]]
        nonZeroCols = numpy.sum(treeSizesDistArray, 0)!=0
        print((Latex.array1DToRow(numpy.arange(treeSizesDistArray.shape[1])[nonZeroCols])))
        print((Latex.array2DToRows(treeSizesDistArray[:, nonZeroCols])))

        treeDepthsDistArray = treeDepthsDistArray[:, 0:treeDepthsDist.shape[0]]
        nonZeroCols = numpy.sum(treeDepthsDistArray, 0)!=0
        print((Latex.array1DToRow(numpy.arange(treeDepthsDistArray.shape[1])[nonZeroCols])))
        print((Latex.array2DToRows(treeDepthsDistArray[:, nonZeroCols])))

        print("Edges and verticies")
        print(Latex.listToRow(dateStrList))
        print(Latex.array2DToRows(numVerticesEdgesArray.T, precision))

def plotOtherStats():

    binEdges = numpy.arange(0, 6000, 180)
    binEdges2 = numpy.arange(0, 10000, 365)
    diffDetections = numpy.zeros((len(subgraphIndicesList2), binEdges.shape[0]-1))
    diffDobs = numpy.zeros((len(subgraphIndicesList2), binEdges2.shape[0]-1))

    global plotInd

    for i in range(len(dayList2)):
        dateStr = (str(DateUtils.getDateStrFromDay(dayList2[i], startYear)))
        logging.info(dateStr)
        subgraph = sGraph.subgraph(subgraphIndicesList2[i])
        subVertexArray = subgraph.getVertexList().getVertices()
        edgeIndices = subgraph.getAllEdges()
        
        diffDetections[i, :], binEdges = numpy.histogram(numpy.abs(subVertexArray[edgeIndices[:, 0], detectionIndex] - subVertexArray[edgeIndices[:, 1], detectionIndex]), binEdges)
        diffDetections[i, :] = diffDetections[i, :]/numpy.sum(diffDetections[i, :])

        diffDobs[i, :], binEdges2 = numpy.histogram(numpy.abs(subVertexArray[edgeIndices[:, 0], dobIndex] - subVertexArray[edgeIndices[:, 1], dobIndex]), binEdges2)
        diffDobs[i, :] = diffDobs[i, :]/numpy.sum(diffDobs[i, :])


        plotInd2 = plotInd 

        plt.figure(plotInd2)
        plt.plot((binEdges[1:]+binEdges[0:-1])/2.0, diffDetections[i, :], label=dateStr)
        plt.xlabel("Difference in detection date")
        plt.ylabel("Probability")
        #plt.ylim((0, 0.8))
        plt.legend()
        plt.savefig(figureDir + "DetectionDatesDist" +  ".eps")
        plotInd2 += 1

        plt.figure(plotInd2)
        plt.plot((binEdges2[1:]+binEdges2[0:-1])/2.0, diffDobs[i, :], label=dateStr)
        plt.xlabel("Difference in DoB")
        plt.ylabel("Probability")
        #plt.ylim((0, 0.8))
        plt.legend()
        plt.savefig(figureDir + "BirthDatesDist" +  ".eps")
        plotInd2 += 1
    
def plotEdgeStats():
    logging.info("Computing vertex stats")

    femaleToHeteroMans = numpy.zeros(len(subgraphIndicesList))
    femaleToBis = numpy.zeros(len(subgraphIndicesList))
    heteroManToFemales = numpy.zeros(len(subgraphIndicesList))
    biToFemales = numpy.zeros(len(subgraphIndicesList))
    biToBis = numpy.zeros(len(subgraphIndicesList))
    
    print(len(subgraphIndicesList))
    print(len(dayList))

    for i in range(len(dayList)):
        logging.info("Date: " + str(DateUtils.getDateStrFromDay(i, startYear)))
        subgraphIndices = subgraphIndicesList[i]
        subgraph = sGraph.subgraph(subgraphIndices)
        subVertexArray = subgraph.getVertexList().getVertices()
        edgeIndices = subgraph.getAllEdges()

        femaleToHeteroMans[i] = numpy.sum(numpy.logical_and(numpy.logical_and(subVertexArray[edgeIndices[:, 0], genderIndex]==1, subVertexArray[edgeIndices[:, 1], genderIndex]==0), subVertexArray[edgeIndices[:, 1], orientationIndex]==0))
        femaleToBis[i] = numpy.sum(numpy.logical_and(numpy.logical_and(subVertexArray[edgeIndices[:, 0], genderIndex]==1, subVertexArray[edgeIndices[:, 1], genderIndex]==0), subVertexArray[edgeIndices[:, 1], orientationIndex]==1))
        heteroManToFemales[i] = numpy.sum(numpy.logical_and(numpy.logical_and(subVertexArray[edgeIndices[:, 0], genderIndex]==0, subVertexArray[edgeIndices[:, 0], orientationIndex]==0), subVertexArray[edgeIndices[:, 1], genderIndex]==1))
        biToFemales[i] = numpy.sum(numpy.logical_and(numpy.logical_and(subVertexArray[edgeIndices[:, 0], genderIndex]==0, subVertexArray[edgeIndices[:, 0], orientationIndex]==1), subVertexArray[edgeIndices[:, 1], genderIndex]==1))
        biToBis[i] = numpy.sum(numpy.logical_and(subVertexArray[edgeIndices[:, 0], orientationIndex]==1, subVertexArray[edgeIndices[:, 1], orientationIndex]==1))

    global plotInd

    plt.figure(plotInd)
    plt.plot(absDayList, femaleToHeteroMans, plotStyles2[0], absDayList, femaleToBis, plotStyles2[1], absDayList, heteroManToFemales, plotStyles2[2], absDayList, biToFemales, plotStyles2[3], absDayList, biToBis, plotStyles2[4])
    plt.xticks(locs, labels)
    plt.xlabel("Year")
    plt.ylabel("Frequency")
    plt.legend(("Woman to Heterosexual Man", "Woman to MSM", "Heterosexual Man to Woman", "MSM to Woman", "MSM to MSM"), loc="upper left")
    plt.savefig(figureDir + "EgoAlterGenderOrient.eps")
    plotInd += 1

#plotTreeStats()
plotScalarStats()
#plotVectorStats()
#plotOtherStats()
#plotEdgeStats()
plt.show()

