import numpy
import os
import zipfile
import pickle
import logging
from apgl.graph.PySparseGraph import PySparseGraph
from apgl.viroscopy.model.HIVVertices import HIVVertices
from apgl.util.Parameter import Parameter

"""
A HIVGraph is a Graph except that its vertices are HIVIndividuals and
there are some useful method to find the S, I and R individuals. 
"""

class HIVGraph(PySparseGraph):
    def __init__(self, numVertices, undirected=True):
        """
        Create a graph with the specified number of vertices, and choose whether
        edges are directed. 
        """
        vList = HIVVertices(numVertices)
        super(HIVGraph, self).__init__(vList, undirected, sizeHint=10000)

    def getSusceptibleSet(self):
        V = self.vList.getVertices(list(range(self.getNumVertices())))
        susceptibleSet = numpy.nonzero(V[:, HIVVertices.stateIndex] == HIVVertices.susceptible)[0]
        return set(susceptibleSet.tolist()) 

    def getInfectedSet(self):
        V = self.vList.getVertices(list(range(self.getNumVertices())))
        infectedSet = numpy.nonzero(V[:, HIVVertices.stateIndex] == HIVVertices.infected)[0]
        return set(infectedSet.tolist())

    def getRemovedSet(self):
        V = self.vList.getVertices(list(range(self.getNumVertices())))
        removedSet = numpy.nonzero(V[:, HIVVertices.stateIndex] == HIVVertices.removed)[0]
        return set(removedSet.tolist())

    def setRandomInfected(self, numInitialInfected, t=0.0):
        Parameter.checkInt(numInitialInfected, 0, self.getNumVertices())
        inds = numpy.random.permutation(numInitialInfected)

        for i in inds[0:numInitialInfected]:
            self.getVertexList().setInfected(i, t)

    @classmethod
    def load(cls, filename):
        """
        Load the graph object from the corresponding file. Tata is loaded in a zip
        format as created using save().

        :param filename: The name of the file to load.
        :type filename: :class:`str`

        :returns: A graph corresponding to the one saved in filename.
        """
        Parameter.checkClass(filename, str)

        (path, filename) = os.path.split(filename)
        os.chdir(path)

        myzip = zipfile.ZipFile(filename + '.zip', 'r')
        myzip.extractall()
        myzip.close()

        matExt = ".mtx"
        W = cls.loadMatrix(filename + matExt)
        vList = VertexList.load(filename)

        boolExt = ".dir"
        file = open(filename + boolExt, 'rb')
        undirected = pickle.load(file)
        file.close()

        graph = cls(vList.getNumVertices(), undirected)
        graph.vList = vList
        graph.W = W

        for tempFile in myzip.namelist():
            os.remove(tempFile)

        logging.info("Loaded epidemic graph from file " + filename)

        return graph
        
