import numpy
import scipy.stats
from pysparse import spmatrix
from apgl.viroscopy.model.HIVVertices import HIVVertices
from apgl.util.Util import *

"""
Model the contact rate of an infected individual and other susceptibles.
"""
class HIVRates():
    def __init__(self, graph):
        self.graph = graph
        self.expandedDegSeq = Util.expandIntArray(graph.outDegreeSequence())
        self.expandedDegSeq = numpy.append(self.expandedDegSeq, numpy.arange(self.graph.getNumVertices()))

        #The hidden degree sequence is taken from a power law distribution, 
        #namely the zipf distribution
        s = 3
        gen = scipy.stats.zipf(s)
        hiddenDegSeq = gen.rvs(size=graph.getNumVertices())
        self.expandedHiddenDegSeq = Util.expandIntArray(hiddenDegSeq)

        self.vList = self.graph.getVertexList()
        self.V = self.vList.getVertices()

        #Parameters for sexual contact
        self.alpha = 2.0
        self.newContactChance = 5.0
        self.heteroContactRate = 0.1
        self.biContactRate = 0.5

        #Infection probabilities are from wikipedia
        #DefaultInfectProb corresponds to two bisexual women
        self.defaultInfectProb = 34.0/10000
        self.womanManInfectProb = 38.0/10000
        self.manWomanHeteroInfectProb = 30.0/10000
        #Taken from "Per-contact probability of HIV transmission in homosexual men in Sydney in the era of HAART" 
        self.manBiInfectProb = 100.0/10000


        #Random detection
        self.randDetectRate = 1/720.0

        #Contact tracing parameters 
        self.ctRatePerPerson = 0.3
        #The start and end time of contact tracing
        self.ctStartTime = 180
        self.ctEndTime = 730

        #Trade off between choosing new contact from known degrees and random (1)
        self.tau = 0.5

        #contactTimesArr is an array of the index of the last sexual contact and contact time
        #Given time starts at zero we set last contact to -inf 
        self.contactTimesArr = numpy.ones((graph.getNumVertices(), 2))*-float('inf')
        self.neighboursList = [] 

        for i in range(graph.getNumVertices()):
            self.neighboursList.append(graph.neighbours(i))

    def setAlpha(self, alpha):
        Parameter.checkFloat(alpha, 0.0, float('inf'))
        self.alpha = alpha

    def setNewContactChance(self, newContactChance):
        Parameter.checkFloat(newContactChance, 0.0, float('inf'))
        self.newContactChance = newContactChance

    def setHeteroContactRate(self, heteroContactRate):
        Parameter.checkFloat(heteroContactRate, 0.0, float('inf'))
        self.heteroContactRate = heteroContactRate

    def setBiContactRate(self, biContactRate):
        Parameter.checkFloat(biContactRate, 0.0, float('inf'))
        self.biContactRate = biContactRate

    def setRandDetectRate(self, randDetectRate):
        Parameter.checkFloat(randDetectRate, 0.0, float('inf'))
        self.randDetectRate = randDetectRate

    def setCtRatePerPerson(self, ctRatePerPerson):
        Parameter.checkFloat(ctRatePerPerson, 0.0, float('inf'))
        self.ctRatePerPerson = ctRatePerPerson

    def setDefaultInfectionProb(self, defaultInfectProb):
        Parameter.checkFloat(defaultInfectProb, 0.0, 1.0)
        self.defaultInfectProb = defaultInfectProb

    def setWomanManInfectProb(self, womanManInfectProb):
        Parameter.checkFloat(womanManInfectProb, 0.0, 1.0)
        self.womanManInfectProb = womanManInfectProb

    def setManWomanHeteroInfectProb(self, manWomanHeteroInfectProb):
        Parameter.checkFloat(manWomanHeteroInfectProb, 0.0, 1.0)
        self.manWomanHeteroInfectProb = manWomanHeteroInfectProb

    def setManBiInfectProb(self, manBiInfectProb):
        Parameter.checkFloat(manBiInfectProb, 0.0, 1.0)
        self.manBiInfectProb = manBiInfectProb

    def contactEvent(self, vertexInd1, vertexInd2, t):
        """
        Indicates a sexual contact event between two vertices. 
        """
        if self.graph.getEdge(vertexInd1, vertexInd2) == None:
            self.expandedDegSeq = numpy.append(self.expandedDegSeq, numpy.array([vertexInd1, vertexInd2]))

        self.graph.addEdge(vertexInd1, vertexInd2, t)
        self.neighboursList[vertexInd1] = self.graph.neighbours(vertexInd1)
        self.neighboursList[vertexInd2] = self.graph.neighbours(vertexInd2)
        self.contactTimesArr[vertexInd1, :] = numpy.array([vertexInd2, t])
        self.contactTimesArr[vertexInd2, :] = numpy.array([vertexInd1, t])

    def removeEvent(self, vertexInd, detectionMethod, t):
        """
        We just remove the vertex from expandedDegSeq and expandedHiddenDegSeq
        """

        self.vList.setDetected(vertexInd, t, detectionMethod)
        self.expandedDegSeq = numpy.delete(self.expandedDegSeq, numpy.nonzero(self.expandedDegSeq==vertexInd))
        self.expandedHiddenDegSeq = numpy.delete(self.expandedHiddenDegSeq, numpy.nonzero(self.expandedHiddenDegSeq==vertexInd))

    def contactRates(self, infectedList, contactList, t):
        """
        Work out contact rates between all infected and non-removed. What is
        computational complexity?
        """
        if len(infectedList) == 0:
            return numpy.array([])

        contactRates = spmatrix.ll_mat(len(infectedList), self.graph.getNumVertices())

        #possibleContacts has as its first column the previous contacts.
        #A set of new contacts based on the degree sequence
        #These contacts can be any one of the other contacts that are non-removed
        possibleContacts = numpy.zeros((len(infectedList), 2), numpy.int)
        possibleContacts[:, 0] = self.contactTimesArr[infectedList, 0]
        
        #Note that we may get duplicates for possible contacts since we don't check for it
        newContacts = numpy.zeros((len(infectedList), 2), numpy.int)
        edsInds = numpy.random.randint(0, self.expandedDegSeq.shape[0], len(infectedList))
        newContacts[:, 0] = self.expandedDegSeq[edsInds]
        hedsInds = numpy.random.randint(0, self.expandedHiddenDegSeq.shape[0], len(infectedList))
        newContacts[:, 1] = self.expandedHiddenDegSeq[hedsInds]
        p = numpy.random.rand(len(infectedList))
        possibleContacts[:, 1] = newContacts[(numpy.arange(len(infectedList)), numpy.array(p<self.tau, numpy.int))]

        #If someone has no previous contact, make sure the probability is zero
        #Could exclude zero probability events from randomChoice if that speed things up
        epsilon = 0.0001
        possibleContactProbs = numpy.zeros((len(infectedList), 2))
        possibleContactProbs[:, 0] = t - self.contactTimesArr[infectedList, 1]
        possibleContactProbs[:, 1] = numpy.ones(len(infectedList))*(2*self.newContactChance)/(self.heteroContactRate + self.biContactRate)
        #Deal with case where last contact time is infinity ago
        infiniteIndices = possibleContactProbs[:, 0]== float('inf')
        possibleContactProbs[:, 0][infiniteIndices] = 1
        possibleContactProbs = (possibleContactProbs+epsilon)**-self.alpha
        possibleContactProbs[:, 0][infiniteIndices] = 0
        #print(possibleContactProbs)

        contactInds = Util.random2Choice(possibleContactProbs).ravel()
        contacts = possibleContacts[(numpy.arange(possibleContacts.shape[0]), contactInds)]

        #We compute sexual contact rates between infected and everyone else
        equalGender = self.V[infectedList, HIVVertices.genderIndex]==self.V[contacts, HIVVertices.genderIndex]
        hInds = numpy.nonzero(numpy.logical_not(equalGender))[0]
        contactRates.put(self.heteroContactRate, hInds, contacts[hInds])

        bInds = numpy.nonzero(numpy.logical_and(numpy.logical_and(equalGender, self.V[contacts, HIVVertices.orientationIndex]==HIVVertices.bi), self.V[infectedList, HIVVertices.orientationIndex]==HIVVertices.bi))[0]
        contactRates.put(self.biContactRate, bInds, contacts[bInds])
        return contactRates

    """
    Compute the infection probability between an infected and susceptible.
    """
    def infectionProbability(self, vertexInd1, vertexInd2, t):
        vertex1 = self.graph.getVertex(vertexInd1)
        vertex2 = self.graph.getVertex(vertexInd2)

        #Remove this? 
        if vertex2[HIVVertices.stateIndex]==HIVVertices.infected or vertex2[HIVVertices.stateIndex]==HIVVertices.removed:
            return 0.0 

        if vertex1[HIVVertices.genderIndex] == HIVVertices.female and vertex2[HIVVertices.genderIndex] == HIVVertices.male:
            return self.womanManInfectProb
        elif vertex1[HIVVertices.genderIndex] == HIVVertices.male and vertex2[HIVVertices.genderIndex] == HIVVertices.female and vertex1[HIVVertices.orientationIndex]==HIVVertices.hetero:
            return self.manWomanHeteroInfectProb
        elif vertex1[HIVVertices.genderIndex] == HIVVertices.male and vertex1[HIVVertices.orientationIndex]==HIVVertices.bi and vertex2[HIVVertices.genderIndex] == HIVVertices.male and vertex2[HIVVertices.orientationIndex]==HIVVertices.bi:
            return self.manBiInfectProb
        else:
            #Corresponds to 2 bisexual women or bisexual male with female 
            return self.defaultInfectProb

    """
    Compute the detection rate of an infected which depends on the entire population.
    """
    def randomDetectionRates(self, infectedList, t):
        detectionRates = numpy.ones(len(infectedList))*self.randDetectRate

        return detectionRates

    def contactTracingRates(self, infectedList, removedSet, t):
        """
        Compute the contact tracing detection rate of a list of infected individuals.
        """
        ctRates = numpy.zeros(len(infectedList))
        #A detected neigbour that happened at ctStartDate or earlier can result in detection
        ctStartDate = t - self.ctStartTime
        #A detected neigbour that happened later than cdEndDate can result in detection
        cdEndDate = t - self.ctEndTime

        infectedArray = numpy.array(infectedList)
        infectedArrInds = numpy.argsort(infectedArray)
        infectedArray = infectedArray[infectedArrInds]

        removeIndices = numpy.array(list(removedSet), numpy.int)
        underCT = numpy.zeros(self.graph.getNumVertices(), numpy.bool)
        underCT[removeIndices] = numpy.logical_and(self.V[removeIndices, HIVVertices.detectionTimeIndex] >= cdEndDate, self.V[removeIndices, HIVVertices.detectionTimeIndex] <= ctStartDate)

        if len(infectedList) < len(removedSet):
            #Possibly store set of detected neighbours
            for i in range(len(infectedList)):
                vertexInd = infectedList[i]
                neighbours = self.neighboursList[vertexInd]

                for ind in neighbours:
                    if underCT[ind]:
                        ctRates[i] += self.ctRatePerPerson
                #This is slower for some reason
                #ctRates[i] = numpy.sum(underCT[neighbours]) * self.ctRatePerPerson
        else:
            for vertexInd in removedSet:
                if underCT[vertexInd]:
                    neighbours = self.neighboursList[vertexInd]

                    for ind in neighbours:
                        if self.V[ind, HIVVertices.stateIndex] == HIVVertices.infected:
                            i = numpy.searchsorted(infectedArray, ind)
                            ctRates[infectedArrInds[i]] += self.ctRatePerPerson
                
        return ctRates